#include "main.h"
#include "tinyfiledialogs.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

#define INIT_SCREEN_WIDTH 800
#define INIT_SCREEN_HEIGHT 600
#define FONT_WIDTH 256
#define FONT_HEIGHT 112		
#define FONT_ROWS 8 		
#define FONT_COLS 32   		
#define FONT_CHAR_W (FONT_WIDTH / FONT_COLS)
#define FONT_CHAR_H (FONT_HEIGHT / FONT_ROWS)
#define FONT_SCALE 1.0
#define TAB_AMT 4
#define UI_Y_SCALE_TOP 2
#define UI_Y_SCALE_BOTTOM 3
#define CURSOR_BLINK_SPEED 0.15
#define FONT_SCALE_PER_WHEEL 0.25
#define SOFT_SCROLL 4096 //if size of line > SOFT_SCROLL we set offset directly to 0 and do not increment it down.
#define MAX_FPS 60
#define TICKS_PER_FRAME 1000 / MAX_FPS
#define GROWTH_FACTOR 1.5f //golden ratio
#define DEFAULT_SIZE 50

#ifdef _WIN32 // we'll need this when testing for linux builds
     #define strdup _strdup
#endif

u64 NOW       	 = 0;
u64 LAST      	 = 0;
float deltaTime  = 0;

Editor E = {0};
GUI    G = {0};
Page*  P = {0};
TILE*  T = {0};

typedef enum {
  FORWARD,
  BACKWARDS
} DIRECTION;

//SYNTAX
typedef struct {
	char* lang;
	char**fileTypes;
	char**keywords;
	char**comments;
  char**strings;
	int flags;
} syntax;

#define SPLITTER "^"

enum syntax_flags {
	NORMAL      = 1 << 0,
	NUMBER      = 1 << 1,
	COMMENT     = 1 << 2,
	STRING      = 1 << 3,
	MATCH       = 1 << 4,
	PRECOMPILED = 1 << 5
};

char* C_keywords[] = {
  "switch", "if", "while", "for", "break", "continue", "return", "else",
  "struct", "union", "typedef", "static", "enum", "class", "case",
  "int", "long", "double", "float", "char", "unsigned", "signed",
  "void", "const", "uint", "u64", "u32", "u16", "u8", "ushort", "uchar", 
  NULL
};
char* C_strings[]  = {SPLITTER, "\"", "\"", "'", "'", NULL};
char* C_files[]    = {".c", ".h", ".cpp", NULL};
char* C_comments[] = {"//", SPLITTER, "/*", "*/", NULL}; //@NOTE strict order of [open, close, open, close] 

syntax SYNTAX[] = {
	{	
		"c",
		C_files,
		C_keywords,
		C_comments,
    C_strings,
		NUMBER | STRING | COMMENT | MATCH | PRECOMPILED
	},
};

enum command_type {
  C_NORMAL,
  C_CONTROL_KEY,
	C_BOTH
};

u8 ctrl_down = 0;

//COMMAND_HELPERS
static void open_file();
static void insertChar(Line* row, char c, ...);
static void save_file();
static void current_syntax(Page* page);
static void insertNewLine();
static void deleteRow();
static void close_page();
static void init_blankpage();
static void p_move(int f);
static void GROW(uint n, uint size, Line* line);
static void deleteChar(Line* row, int at, DIRECTION dir);
static int IN_RANGE(char x) {return ((x < 65 || x > 94) && (x < 97 || x > 122));}
static int get_word_pos(Line* line, uint at, u8 forward) {
	if(forward == 0){
		if(at < 1) return;
		for(uint i = at - 1; i >= 0; i--) {
			if(i == 0 || IN_RANGE(line->chars[i])) return i;
		}
	} else {
		if(at >= line->size) return;
		for(uint i = at + 1; i <= line->size; i++) {
			if(IN_RANGE(line->chars[i]) || i == line->size) return i;
		}
	}
}

//COMMANDS
static void PAGE_DOWN(){ if(E.buffer_cursor.y < P->lines) E.buffer_cursor.y = (E.buffer_cursor.y + E.visibleH > P->lines) ? E.buffer_cursor.y = P->lines - 1 : E.buffer_cursor.y + E.visibleH; }
static void PAGE_UP(){ E.buffer_cursor.y = (E.buffer_cursor.y > E.visibleH) ? (E.buffer_cursor.y - E.visibleH) - 1 : 0; }	
static void CURSOR_UP(){ 
	if(E.buffer_cursor.y > 0) {
		if(ctrl_down) PAGE_UP();
		else {
      E.buffer_cursor.y--;
      if(E.buffer_cursor.x > P->buffer[E.buffer_cursor.y].size) E.buffer_cursor.x = P->buffer[E.buffer_cursor.y].size;
    }
	}
}
static void CURSOR_DOWN() {
	if(ctrl_down) PAGE_DOWN();
	else {
		if(E.buffer_cursor.y < P->lines - 1) {
      E.buffer_cursor.y++;
      if(E.buffer_cursor.x > P->buffer[E.buffer_cursor.y].size) E.buffer_cursor.x = P->buffer[E.buffer_cursor.y].size;
    }
		//if(E.buffer_cursor.y > P->lines - 1) insertNewLine(); //maybe not the best idea?
	}
}
static void CURSOR_RIGHT()  {
  if(E.buffer_cursor.x >= P->buffer[E.buffer_cursor.y].size) {
    E.buffer_cursor.x = 0;
    CURSOR_DOWN();
  } else E.buffer_cursor.x++;
}
static void CURSOR_LEFT()   {
  if(E.buffer_cursor.x == 0 && E.buffer_cursor.y > 0){
    E.buffer_cursor.y--;
    E.buffer_cursor.x = P->buffer[E.buffer_cursor.y].size;
  } else if(E.buffer_cursor.x > 0) E.buffer_cursor.x--;
}
static void RETURN() {
	E.buffer_cursor.y++;
	insertNewLine();
	E.buffer_cursor.x = 0; //Might be cool to do alignment to the beginning of first char that isn't a space?
}
static void TAB() { //more efficient if we just inserted '\t'
	if(E.buffer_cursor.x < P->buffer[E.buffer_cursor.y].size) {
		int at = E.buffer_cursor.x;
		Line* line = &P->buffer[E.buffer_cursor.y];
		GROW(line->size + TAB_AMT + 1, line->s_a, line);
		memmove(&line->chars[at + TAB_AMT], &line->chars[at], line->size - at + TAB_AMT);
		for(int i = 0; i < TAB_AMT; i++) {
			line->chars[at + i] = 32;
		}
		line->size += TAB_AMT;
		line->chars[line->size] = '\0';
	}
}
static void R_CTRL()      { ctrl_down = 1; }
static void L_CTRL()      { ctrl_down = 1; }
static void END_LINE()    { if(ctrl_down == 1) if(P->buffer[E.buffer_cursor.y].size > 0)E.buffer_cursor.x = P->buffer[E.buffer_cursor.y].size; }
static void START_LINE()  {
  static uint true_first = 0;
  Line* line = &P->buffer[E.buffer_cursor.y];
  if(ctrl_down == 1) {
    if(true_first > 0) {
      E.buffer_cursor.x = 0;
      true_first = 0;
      return;
    }
    for(uint i = 0; i < line->size; i++){
      if(line->chars[i] > 32 && line->chars[i] < 127){
        E.buffer_cursor.x = i;
        break;
      }
    }
    true_first+=1;
  }
}
static void DELETE_WORD_B() {	
	Line* line = &P->buffer[E.buffer_cursor.y];
	uint i = get_word_pos(line, E.buffer_cursor.x, 0);
	memmove(&line->chars[i], &line->chars[E.buffer_cursor.x], line->size - E.buffer_cursor.x);
	line->size -= (E.buffer_cursor.x - i);
	E.buffer_cursor.x = line->size;
}
static void DELETE_WORD_F() {
  Line* line = &P->buffer[E.buffer_cursor.y];
  uint i  = get_word_pos(line, E.buffer_cursor.x, -1);
  memmove(&line->chars[E.buffer_cursor.x], &line->chars[i], line->size - E.buffer_cursor.x);
	line->size -= (i - E.buffer_cursor.x);
}
static void BACKSPACE() {
	if(P->buffer[E.buffer_cursor.y].size > 0) {
		if(E.buffer_cursor.x == 0) {
			if(P->lines < 1 || E.buffer_cursor.y == 0) return;
			Line* curr_line = &P->buffer[E.buffer_cursor.y];
			Line* line 		= &P->buffer[E.buffer_cursor.y - 1];
			GROW(line->size + curr_line->size, line->s_a, line);
			memmove(&line->chars[line->size], &curr_line->chars[0], curr_line->size);
			line->size += curr_line->size;
			line->chars[line->size + 1] = '\0';
			deleteRow(BACKWARDS);
		} else {
			deleteChar(&P->buffer[E.buffer_cursor.y], E.buffer_cursor.x, BACKWARDS);
    }						
	} else if(P->lines > 1) {
		deleteRow(BACKWARDS);
	}
}
static void BACKSPACE_F() {
  if(P->buffer[E.buffer_cursor.y].size > 0 && E.buffer_cursor.x < P->buffer[E.buffer_cursor.y].size) {
    deleteChar(&P->buffer[E.buffer_cursor.y], E.buffer_cursor.x, FORWARD);
  } else if(P->lines > 1) {
    deleteRow(FORWARD);
  }
}
static void NEXT_WORD() { E.buffer_cursor.x = get_word_pos(&P->buffer[E.buffer_cursor.y], E.buffer_cursor.x, 1); }
static void LAST_WORD() { E.buffer_cursor.x = get_word_pos(&P->buffer[E.buffer_cursor.y], E.buffer_cursor.x, 0); }
static void MARK(){
	if(E.mark.x == E.buffer_cursor.x && E.mark.y == E.buffer_cursor.y) E.mark = vec2is(-1);
  else E.mark = vec2i(E.buffer_cursor.x, E.buffer_cursor.y);
}
static void LAST_PAGE() { p_move(-1);  }
static void NEXT_PAGE() { p_move(1); }
static void NEW_PAGE() {
	/* E.pages    		= realloc(E.pages, sizeof(Page) * (E.p_total + 1)); */
	T->p_total 	   += 1;
	T->p_pos        = T->p_total - 1;
	P = &T->pages[T->p_pos];
	init_blankpage();
  KW_SetLabelText(G.footer->children[0], P->full[0] != '\0' ? P->full : "empty buffer");
}

//Commands table
//Eventually we could add a config to change the key mappings
Command command_table[] = {
  { SDLK_RIGHT, &CURSOR_RIGHT, C_NORMAL, 0 },
	{ SDLK_LEFT, &CURSOR_LEFT, C_NORMAL, 1 },
	{ SDLK_UP, &CURSOR_UP, C_BOTH, 2},
	{ SDLK_DOWN, &CURSOR_DOWN, C_BOTH, 3},
	{ SDLK_PAGEUP, &PAGE_UP, C_NORMAL, 4},
	{ SDLK_PAGEDOWN, &PAGE_DOWN, C_NORMAL, 5},
	{ SDLK_RETURN, &RETURN, C_NORMAL, 6},
	{ SDLK_TAB, &TAB, C_NORMAL, 7},
	{ SDLK_RCTRL, &R_CTRL, C_NORMAL, 8},
	{ SDLK_LCTRL, &L_CTRL, C_NORMAL, 9},
	{ SDLK_BACKSPACE, &BACKSPACE, C_NORMAL, 10},
	{ SDLK_a, &START_LINE, C_CONTROL_KEY, 11},
	{ SDLK_e, &END_LINE, C_CONTROL_KEY, 12},
  { SDLK_DELETE, &DELETE_WORD_F, C_CONTROL_KEY, 13},
	{ SDLK_BACKSPACE, &DELETE_WORD_B, C_CONTROL_KEY, 14},
	{ SDLK_LEFT, &LAST_WORD, C_CONTROL_KEY, 15},
	{ SDLK_RIGHT, &NEXT_WORD, C_CONTROL_KEY, 16},
	{ SDLK_o, &open_file, C_CONTROL_KEY, 17},
	{ SDLK_SPACE, &MARK, C_CONTROL_KEY, 18},
	{ SDLK_n, &NEXT_PAGE, C_CONTROL_KEY, 19},
	{ SDLK_b, &LAST_PAGE, C_CONTROL_KEY, 20},
	{ SDLK_x, &close_page, C_CONTROL_KEY, 21},
	{ SDLK_w, &NEW_PAGE, C_CONTROL_KEY, 22},
  { SDLK_s, &save_file, C_CONTROL_KEY, 23},
  { SDLK_DELETE, &BACKSPACE_F, C_NORMAL, 24},
};

//Hex to R G B A
#define UNHEXA(x)\
  ((x) >> (8 * 3)) & 0xFF, \ 
  ((x) >> (8 * 2)) & 0xFF, \ 
  ((x) >> (8 * 1)) & 0xFF, \ 
  ((x) >> (8 * 0)) & 0xFF    

//Hex to R G B
#define UNHEX(x)\
  ((x) >> (8 * 2)) & 0xFF, \ 
  ((x) >> (8 * 1)) & 0xFF, \ 
  ((x) >> (8 * 0)) & 0xFF


//ERROR HANDLES SDL
  static void die(int code) {
	if(code < 0) {
		fprintf(stderr,"SDL ERROR %s\n", SDL_GetError());
		exit(1);
	}
}

static void* die_ptr(void *ptr) {
	if(ptr == NULL) {
		fprintf(stderr,"SDL ERROR %s\n", SDL_GetError());
		exit(1);
	}
	return ptr;
}

//FONT LOADING
static SDL_Surface* surface_from_file(const char* path) {
	int x,y,n;
	unsigned char* pixels = stbi_load(path, &x, &y, &n, STBI_rgb_alpha);
	if(pixels == NULL) {
		fprintf(stderr, "ERROR: could not load file %s:  %s\n", path, stbi_failure_reason());
		exit(1);
	}

  //This code is taken directly from sdl's documentation
	u32 rmask, gmask, bmask, amask;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	rmask = 0xff000000;
	gmask = 0x00ff0000;
	bmask = 0x0000ff00;
	amask = 0x000000ff;
#else
	rmask = 0x000000ff;
	gmask = 0x0000ff00;
	bmask = 0x00ff0000;
	amask = 0xff000000;
#endif

	const int depth = 32;
	const int pitch = 4 * x;

	return die_ptr(SDL_CreateRGBSurfaceFrom((void*)pixels, x, y, depth, pitch, rmask, gmask, bmask, amask)); 
}

static Font font_load_from_file(const char* path, SDL_Renderer* renderer) {
	Font font = {0};
	SDL_Surface* font_surface = surface_from_file(path);
	font.charMap = die_ptr(SDL_CreateTextureFromSurface(renderer, font_surface));
	SDL_FreeSurface(font_surface);

	for(uint ascii = ASCII_DISPLAY_LOW; ascii <=  ASCII_DISPLAY_HIGH; ++ascii) {
		const uint idx = ascii - ASCII_DISPLAY_LOW;
		const uint col = idx % FONT_COLS;
		const uint row = idx / FONT_COLS;

		font.glyph_table[idx] = (SDL_Rect){
			.x = col * FONT_CHAR_W,
			.y = row * FONT_CHAR_H,
			.w = FONT_CHAR_W,
			.h = FONT_CHAR_H
		};
	}

	return font;
}

//MEMORY MANAGEMENT
static void GROW(uint n, uint size, Line* line) {
	if(n >= size) {
		line->s_a 	= (int)size * GROWTH_FACTOR;
		line->chars = realloc(line->chars, line->s_a);
	}
}
static void GROW_B(uint n, uint size) {
	if(n >= size) {
		P->s_a 		= (int)size * GROWTH_FACTOR;
		P->buffer = realloc(P->buffer, (P->s_a * sizeof(Line)));
	}
}

static void SHRINK(uint n, uint size, Line* line) {
	if(size == DEFAULT_SIZE) return;
  int m = (uint)ceilf(size / GROWTH_FACTOR);
	if(n < m) {
		line->s_a 	= m;
		line->chars = realloc(line->chars, line->s_a);
	}
}
static void SHRINK_B(uint n, uint size) {
	if(size == DEFAULT_SIZE) return;
  int m = (int)ceilf(size / GROWTH_FACTOR);
  if(n < m) {
    P->s_a    = m;
    P->buffer = realloc(P->buffer, P->s_a * sizeof(Line));
  }
}

//TILE
static void init_default_tile(){ //call this only on startup as we want one default tile
  T->t_size  = uvec2is(0);
  T->p_pos   = 0;
  T->p_total = 1;
  P = &T->pages[T->p_pos];
  init_blankpage();
}

static void init_new_tile() {
  E.t_total      += 1;
  E.current_tile += 1;
  E.tiles         = realloc(E.tiles, sizeof(TILE) * E.t_total);

  TILE z = {0};
  z.t_size  = uvec2is(0); //this should change based on the current # of tiles open for testing eh fuck it
  z.p_pos   = 0;
  z.p_total = 1;
  E.tiles[E.current_tile] = z;
  T = &E.tiles[E.current_tile];
  P = &T->pages[T->p_pos];
  init_blankpage();
}


//SCROLLING
static void scroll() {
	if(E.buffer_cursor.x < E.offset.x) E.offset.x = 0;//(E.offset.x > 0 && E.offset.x < SOFT_SCROLL) ? E.offset.x - 1 : 0;
	if(E.buffer_cursor.x >= E.offset.x + E.visibleW) E.offset.x = E.buffer_cursor.x - E.visibleW;
	if(E.buffer_cursor.y < E.offset.y) E.offset.y = 0;
	if(E.buffer_cursor.y >= E.offset.y + E.visibleH) E.offset.y = E.buffer_cursor.y - E.visibleH;
}

//SYNTAX
static u32 syntaxToColor(uint syn, u32 DEFAULT_COLOR) {
	switch(syn) {
  case MATCH: return 0x199EDE; 		    //blue (25, 158, 222)
  case COMMENT: return 0x7d7d7d; 		  //light blue (0, 183, 255)
  case NUMBER: return 0x00b7ff;		    //gray (125, 125, 125) 
  case STRING: return 0xDE199E; 		  //pink (222, 25, 158)
  case PRECOMPILED: return 0x0C4F6F;  //dblue (12, 79, 111)
  default: return DEFAULT_COLOR;
	}
}

static void current_syntax(Page* page) {
	for(uint i = 0; i < sizeof(SYNTAX) / sizeof(SYNTAX[0]); i++) {
		char** t = SYNTAX[i].fileTypes;
		for(char* z = *t; z; z = *++t) {
			if(strstr(page->f_type, z) != NULL) return page->c_syntax = i; 
		}
	}
	return page->c_syntax = -1;
}

static int IN_R(char a) { return (a < 65 && (a < 48 || a > 57) || a > 122); }
static int check(uint start, uint end, Line line, uint type) {
  if(type & MATCH) return (start == 0 || IN_R(line.chars[start-1])) && (IN_R(line.chars[end+1]) || end+1 == line.size);
  if(type & COMMENT) return (start == 0 || (line.chars[start-1] != '"' && line.chars[start-1] != '\''));
  return 1;
}
static uint check_for(char** arr, uint flag, Line line, PQ* q, uint c_line, u8 multi) {
  char* x         = line.chars;
  static int is_c = -1; //is multiline closed
  static char*o_s = ""; //what is the multi line start char
  static uint l_s = 0;  //line where multi line inserted
  uint start      = 0;
  uint end        = line.size - 1;
  uint c          = 0; //hit splitter for multi line checking now
  for(char* i = *arr; i != NULL; i = *++arr) {
    x = line.chars;
    if((i[0] == SPLITTER[0] && i[1] == '\0')) { //IF SPLITTER HIT
      if(multi > 0) c = 1;
    } else if(c > 0) { //MULTI LINE CHECKING
      char* z = i;
      if(is_c > -1 && l_s <= c_line) { //check all closings to see if found on line
        z = *++arr; //get the closing char
        if(z == NULL) break;
        start = 0;
        end   = line.size - 1;
        if((x = strstr(x, z)) != NULL) {
          if(strcmp(o_s, *--arr) == 0) { //cmp if opening char is equal to the found closing char's opening char
            end = (x - line.chars) + strlen(x) - 1;
            is_c = -1;
          }
        }
        pq_push(q, init_n((Vec3i){ .x = start, .y = end, .z = flag })); //push to queue
      } else {
        //get opening char and check if closing exist on that line
        u8 found = 0;
        char* closing_char = *++arr;
        while((x = strstr(x, i)) != NULL) {
          found        = 0;
          start        = x - line.chars;
          end          = 0;
          x           += strlen(i);
          if(closing_char == NULL) break; //no closing char in array... bad
          if((x = strstr(x, closing_char)) != NULL) {
            end = (flag & STRING) ? (x - line.chars): (x - line.chars) + strlen(x) - 1;
            x += strlen(closing_char);
            found = 1;
          }
          if(check(start, end, line, flag)) {
            if(pq_push(q, init_n((Vec3i) { .x = start, .y = end, .z = flag })) == 0 && (flag & COMMENT) && found == 0) {
              is_c = 1;
              l_s  = c_line;
              o_s  = *--arr;
              break;
            }
          }
          if(x == NULL) break;
        }
        i = *++arr;
        if(i == NULL) break;
      }
    } else { //SINGLE LINE
      while((x = strstr(x, i)) != NULL) { //we need to search for all instances in regards to match
        uint len   = strlen(i);
        start	     = x - line.chars;
        end        = (flag & MATCH) ? start + len - 1 : line.size - 1;
        if(check(start, end, line, flag)) {
          pq_push(q, init_n((Vec3i){ .x = start, .y = end, .z = flag }));
        }
        x += len;
      }
    }
  }
  return start;
}
static void highlight(uint max, Line* line, PQ* q, syntax lang, uint l) {
	if(!lang.lang || line->size == 0) return;
	uint start = 0;
  char* x = line->chars;
	if(lang.flags & COMMENT)
    start = check_for(lang.comments, COMMENT, *line, q, l, 1);
	if(lang.flags & STRING)
    start = check_for(lang.strings, STRING, *line, q, l, 1);
	if(lang.flags & PRECOMPILED)
    if(x[0] == '#') pq_push(q, init_n((Vec3i){ .x = 0, .y = (start > 0) ? start - 1: line->size - 1, .z = PRECOMPILED}));
	if(lang.flags & MATCH)
    start = check_for(lang.keywords, MATCH, *line, q, l, 0);
}

//RENDERING
static void render_char(SDL_Renderer* renderer, Font* font, Vec2f pos, u32 color, float scale, u8 c) {
	const SDL_Rect dst = {
		.x = pos.x,
		.y = pos.y,
		.w = FONT_CHAR_W * scale,
		.h = FONT_CHAR_H * scale
	};
	uint index = c - ASCII_DISPLAY_LOW;
	if(c > ASCII_DISPLAY_HIGH || c < ASCII_DISPLAY_LOW) index = 0; //ERROR CHARACTER replaced with space
	die(SDL_RenderCopy(renderer, font->charMap, &font->glyph_table[index], &dst));
}

static void render_text_sized(SDL_Renderer* renderer, Font* font, Vec2f pos, u32 color, float scale, Line* buffer) {
	die(SDL_SetTextureAlphaMod(font->charMap, 0xFF));
	Vec2f pen = pos;
	uint max_y = (P->lines < E.visibleH && E.visibleH > 0) ? P->lines : E.visibleH + E.offset.y + 1;
	uint max_x = 0;
	PQ* syn = init_pq();
	Vec3i t = vec3is(-1);
	for(uint y = E.offset.y; y < max_y; y++) {
    if(y >= P->lines) break;
		pen.x = pos.x;
    max_x = ((P->buffer[y].size <= E.visibleW) && E.visibleW > 0) ? P->buffer[y].size : E.visibleW + E.offset.x;
    highlight(max_x, &P->buffer[y], syn, SYNTAX[P->c_syntax], y);
		for(uint x = E.offset.x; x < max_x; x++) {
      if(x >= P->buffer[y].size) break;
      if(x > t.y || t.y == -1) t = pq_pop(syn);
      if(x >= t.x && x <= t.y) die(SDL_SetTextureColorMod(font->charMap, UNHEX(syntaxToColor(t.z, color))));
      else die(SDL_SetTextureColorMod(font->charMap, UNHEX(color)));
      render_char(renderer, font, pen, color, scale, buffer[y].chars[x]);
			pen.x += FONT_CHAR_W * scale;
		}
		t = vec3is(-1);
		pq_empty(syn);
		pen.y += FONT_CHAR_H * scale;
	}
}

//GENERAL EDITOR
static void insertRow(const char* text, int at, uint len) {
	if(at < 0 || at > P->lines) return;

	//Reallocate the size of the buffer to account for a new line!
  P->lines += 1;
	GROW_B(P->lines, P->s_a);
  memmove(&P->buffer[at + 1], &P->buffer[at], sizeof(Line) * (P->lines - at));

	//Allocate text to the chars buffer
	P->buffer[at].size  = len;
	P->buffer[at].s_a   = (len > DEFAULT_SIZE) ? len * GROWTH_FACTOR : DEFAULT_SIZE;
	P->buffer[at].chars = malloc(P->buffer[at].s_a);
	memcpy(P->buffer[at].chars, text, len);
	P->buffer[at].chars[len] = '\0';
}

//We don't want a ton of random spaces in the line especially after the end character
//So we trim it, this probably could be enabled or disabled feature but it does help keep the lines shorter
static void trimLine(Line* line) {	
	for(uint i = line->size - 1; i > 0; i--) {
		if(line->chars[i] != 32) return;
		line->size--;
		SHRINK(line->size, line->s_a, line);
	}
}

static void insertNewLine() {
	if(P->buffer[E.buffer_cursor.y - 1].size == 0 || E.buffer_cursor.x == P->buffer[E.buffer_cursor.y - 1].size) {
		insertRow("", E.buffer_cursor.y, 0);
		E.buffer_cursor.x = 0;
	}
	else {		
		Line* line = &P->buffer[E.buffer_cursor.y - 1];
		insertRow(&line->chars[E.buffer_cursor.x], E.buffer_cursor.y, line->size - E.buffer_cursor.x);
		line = &P->buffer[E.buffer_cursor.y - 1];
		line->size = E.buffer_cursor.x;
		line->chars[line->size] = '\0';
	}
	/* P->dirty += 1; */
}

static void deleteRow(DIRECTION dir) {
  free(P->buffer[E.buffer_cursor.y].chars);
	memmove(&P->buffer[E.buffer_cursor.y], &P->buffer[E.buffer_cursor.y + 1],  sizeof(P->buffer[0]) * (P->lines - E.buffer_cursor.y - 1)); //If line below our removal we move it up
	if(dir == BACKWARDS && E.buffer_cursor.y > 0) {
    E.buffer_cursor.x = P->buffer[E.buffer_cursor.y - 1].size;
    E.buffer_cursor.y--;
  }	else if(E.buffer_cursor.y + 1 < P->lines) E.buffer_cursor.x = P->buffer[E.buffer_cursor.y + 1].size;
  P->lines -= 1;
	SHRINK_B(P->lines, P->s_a);
  P->dirty = 1;
}


static void deleteChar(Line* row, int at, DIRECTION dir) {
	if(dir == BACKWARDS) {
    if(E.buffer_cursor.x == 0) return;
    memmove(&row->chars[at - 1], &row->chars[at], row->size - at);
    E.buffer_cursor.x--;
  } else memmove(&row->chars[at], &row->chars[at+1], row->size - at);
	row->size--;
	SHRINK(row->size, row->s_a, row);
	P->dirty += 1;
}

static void insertLine(Line* row, const char* str, uint len) { //only used right now for files as of now
  if(row == NULL || len == 0) return;
  /* if(len > row->s_a) { */
    row->s_a += len;
    row->chars = realloc(row->chars, row->s_a);
  /* } */
    //	memcpy(row->chars + row->size - 1, str, len);
    strcat(row->chars, str);
	row->size += len;
	//row->chars[row->size] = '\0';
}

static void insertChar(Line* row, char c, ...) {
	va_list args;
	va_start(args, c);

	int CLOSING_CHAR = va_arg(args, int);
	int at = E.buffer_cursor.x;
	GROW(row->size + 1, row->s_a, row);
	memmove(&row->chars[at+1], &row->chars[at], row->size - at + 1);
	row->size++;

	row->chars[at] = c;
	row->chars[row->size] = '\0';

	P->dirty += 1;

	if(CLOSING_CHAR != -1) E.buffer_cursor.x++;

#if DEBUG //@TODO this should be a feature in the config
	if(P->c_syntax > -1) {
		if(c == '{' || c == '(' || c == '[') insertChar(row, (c == '(') ? c + 1 : c + 2, -1); 
		else if((c == 34 || c == 39) && CLOSING_CHAR != -1) insertChar(row, c, -1); //Chars like ' and "
	}
#endif
	va_end(args);
}
static void render_cursor(SDL_Renderer* renderer, u32 color, int a, Vec2f pos) {
	SDL_Rect rect = {
		.x = ((E.buffer_cursor.x - E.offset.x) * FONT_CHAR_W * E.font_scale) + pos.x,
		.y = ((E.buffer_cursor.y - E.offset.y) * FONT_CHAR_H * E.font_scale) + pos.y,
		.w = FONT_CHAR_W * E.font_scale,
    .h = FONT_CHAR_H * E.font_scale
	};

	die(SDL_SetRenderDrawColor(renderer, UNHEX(color), a));
	SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_BLEND);
	die(SDL_RenderFillRect(renderer, &rect));
}

static void render_mark(SDL_Renderer* renderer, u32 color, int a, Vec2f pos) {
  if((E.mark.x == -1 || E.mark.y == -1)  && (E.buffer_cursor.x > E.mark.x && E.buffer_cursor.y > E.mark.y)) return;
  uint x = (E.buffer_cursor.y > E.mark.y) ? E.visibleW : E.buffer_cursor.x;

  SDL_Rect rect = {
    .x = (E.mark.x * FONT_CHAR_W * E.font_scale) + pos.x,
    .y = (E.mark.y * FONT_CHAR_H * E.font_scale) + pos.y,
    .w =  x * FONT_CHAR_W * E.font_scale,
    .h = (E.buffer_cursor.y + 1) * FONT_CHAR_H * E.font_scale,
  };

  die(SDL_SetRenderDrawColor(renderer, UNHEX(color), a));
	SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_BLEND);
	die(SDL_RenderFillRect(renderer, &rect));
}

//GUI
KW_Widget* file_dropdown = NULL;
KW_Rect dropdownGeo      = {0};

static void close_dropdown() {
	KW_HideWidget(file_dropdown);
}

static void show_dropdown(KW_Widget* widget) {
	KW_SetWidgetGeometry(file_dropdown, &dropdownGeo);
	KW_ShowWidget(file_dropdown);
}

static void init_gui(SDL_Window* window, SDL_Renderer* renderer) {
	G.driver 			= KW_CreateSDL2RenderDriver(renderer, window);
	G.set 	 			= KW_LoadSurface(G.driver, "tileset-alloy.png");
  G.font        = KW_LoadFont(G.driver, "Perfect DOS VGA 437.ttf", 12);
	G.gui 	 			= KW_Init(G.driver, G.set);

  KW_SetFont(G.gui, G.font);

	G.header_geo = (KW_Rect) {
    .x = 0,
    .y = 0,
    .w = INIT_SCREEN_WIDTH,
    .h = ((FONT_CHAR_H * FONT_SCALE) * 2)
  };
	G.header		 	      = KW_CreateFrame(G.gui, NULL, &G.header_geo);
	//KW_Widget * label 	= KW_CreateLabel(G.gui, G.header, "AZUL MEGAMIND GUI", &G.header_geo);
	G.header_geo.w 	  	= G.header_geo.w / 10;
	KW_Widget* btn 	  	= KW_CreateButtonAndLabel(G.gui, G.header, "FILE", &G.header_geo);

	dropdownGeo = (KW_Rect) {
    .x = 0,
    .y = ((FONT_CHAR_H * FONT_SCALE) * 2),
    .w = INIT_SCREEN_WIDTH / 10,
    .h = 120
  };
	file_dropdown 		  = KW_CreateFrame(G.gui, btn, &dropdownGeo);
	KW_Widget* new_file = KW_CreateButtonAndLabel(G.gui, file_dropdown, "NEW", &(KW_Rect){
      .x = 0,
      .y = 0,
      .w = INIT_SCREEN_WIDTH / 10,
      .h = 40
    });
	KW_Widget* file_opn = KW_CreateButtonAndLabel(G.gui, file_dropdown, "OPEN", &(KW_Rect){
      .x = 0,
      .y = 40,
      .w = INIT_SCREEN_WIDTH / 10,
      .h = 40
    });

	KW_Widget* save_file = KW_CreateButtonAndLabel(G.gui, file_dropdown, "SAVE", &(KW_Rect){
      .x = 0,
      .y = 80,
      .w = INIT_SCREEN_WIDTH / 10,
      .h = 40
    });

	G.footer_geo = (KW_Rect) {
    .x = 0,
    .y = INIT_SCREEN_WIDTH - ((FONT_CHAR_H * FONT_SCALE) * 2),
    .w = INIT_SCREEN_WIDTH,
    .h = ((FONT_CHAR_H * FONT_SCALE) * 2)
  };
	G.footer 			= KW_CreateFrame(G.gui, NULL, &G.footer_geo);
  //need outline around page so show selected
	KW_Widget* f_name = KW_CreateLabel(G.gui, G.footer, "empty buffer" , &(KW_Rect){
      .x = 0,
      .y = 0,
      .w = G.footer_geo.w,
      .h = G.footer_geo.h
    });
  KW_Widget* f_line = KW_CreateLabel(G.gui, G.footer, "(0, 0)", &(KW_Rect){
      .x = G.footer_geo.w / 2,
      .y = 0,
      .w = G.footer_geo.w,
      .h = G.footer_geo.h
    });

	KW_SetLabelTextColor(f_name, (KW_Color){.r = 255, .g = 255, .b = 255, .a = 255});
  KW_SetLabelTextColor(f_line, (KW_Color){.r = 255, .g = 255, .b = 255, .a = 255});
	KW_SetLabelAlignment(f_name, KW_LABEL_ALIGN_LEFT, 10, KW_LABEL_ALIGN_MIDDLE, 0);
  KW_SetLabelAlignment(f_line, KW_LABEL_ALIGN_CENTER, 0, KW_LABEL_ALIGN_MIDDLE, 0);
	KW_EnableWidgetDebug(G.header, 1);

	KW_HideWidget(file_dropdown);
	KW_AddWidgetMouseDownHandler(btn, show_dropdown);
	KW_AddWidgetMouseLeaveHandler(file_dropdown, close_dropdown);
	KW_AddWidgetMouseDownHandler(file_opn, open_file);
  KW_AddWidgetMouseDownHandler(save_file, save_file);
  KW_AddWidgetMouseDownHandler(new_file, init_blankpage);
}

static void update_gui() {
	G.header_geo.w = E.screenW;
	G.footer_geo.y = E.screenH - ((FONT_CHAR_H * FONT_SCALE) * 2);
	G.footer_geo.w = E.screenW;
	KW_SetWidgetGeometry(G.header, &G.header_geo);
  KW_SetWidgetGeometry(G.footer, &G.footer_geo);
}

static void render_gui() {
	KW_ProcessEvents(G.gui);
	KW_Paint(G.gui);
}

//PAGE
static void p_move(int f) {
	if(T->p_pos == T->p_total - 1 && f > 0) return;
	else if(T->p_pos == 0 && f < 0) return;
	P->l_off 		= E.offset;
	P->l_pos 		= E.buffer_cursor;
	if(f > 0)  T->p_pos += 1;
	else T->p_pos -= 1;
	P = &T->pages[T->p_pos];
	E.buffer_cursor = P->l_pos;
	E.offset		    = P->l_off;
  KW_SetLabelText(G.footer->children[0], (P->full[0] != '\0' ? P->full : "empty buffer"));
  E.mark     = vec2is(-1);
}

static void free_page(Page* f) {
	free(f->buffer);
	free(f->f_type);
	free(f->f_name);
	free(f->full);
  free(f->path);
}

//current page blank
static void init_blankpage() {
	P->lines  	   = 0;
	P->buffer 	   = malloc(DEFAULT_SIZE * sizeof(Line));
	P->s_a	       = DEFAULT_SIZE;
	P->f_type      = malloc(1);
	P->f_name 	   = malloc(1);
  P->path        = malloc(1);
	P->full		     = malloc(1);
	P->full[0]	   = '\0';
	P->f_type[0]   = '\0';
	P->f_name[0]   = '\0';
  P->path[0]     = '\0';
	P->c_syntax    = -1;
	P->l_pos	     = uvec2is(0);
	P->l_off	     = uvec2is(0);
	P->dirty	     = 0;

	E.buffer_cursor = P->l_pos;
	E.offset 		    = P->l_off;

  insertRow("", P->lines, 0);
}

//this is a new page
static void init_page(const char* f_name, const char* f_type, const char* full, const char* path) {
	Page page = {0};
	uint n_l = strlen(f_name), t_l = strlen(f_type), p_l = strlen(path);
	page.lines 		 = 0;
	page.s_a   		 = DEFAULT_SIZE;
	page.buffer 	 = malloc(DEFAULT_SIZE * sizeof(Line));
	page.f_type 	 = malloc(t_l);
	page.f_name 	 = malloc(n_l);
	page.full		   = malloc(n_l + t_l);
  page.path      = malloc(p_l);
	memmove(page.f_type, f_type, t_l);
	memmove(page.f_name, f_name, n_l);
	memmove(page.full,   full, t_l + n_l);
  strcpy(page.path, path);
	page.l_pos		 	     = uvec2is(0);
	page.l_off		 	     = uvec2is(0);
	page.f_type[t_l] 	   = '\0';
	page.f_name[n_l] 	   = '\0';
	page.full[n_l + t_l] = '\0';
  page.path[p_l]       = '\0';
  
	current_syntax(&page);
  T->p_total        += 1;
	T->p_pos           = T->p_total - 1;
	T->pages[T->p_pos] = page;
	P                  = &T->pages[T->p_pos];
}

//function is a half truth, when calling this we only ever edit the specific values passed in.
static void edit_page(const char* f_name, const char* f_type, const char* full, const char* path) {
  uint n_l = strlen(f_name), t_l = strlen(f_type), p_l = strlen(path);
  P->f_type = realloc(P->f_type, t_l);
  P->f_name = realloc(P->f_name, n_l);
  P->full   = realloc(P->full, n_l + t_l);
  P->path   = realloc(P->path, p_l);
  memcpy(P->f_type, f_type, t_l);
  memcpy(P->f_name, f_name, n_l);
  memcpy(P->path, path, p_l);
  memcpy(P->full, full, n_l + t_l);
  P->f_type[t_l]     = '\0';
  P->f_name[n_l]     = '\0';
  P->full[n_l + t_l] = '\0';
  P->path[p_l]       = '\0';
}

static void close_page() {
	if(T->p_total == 1) {
		free_page(P);
		init_blankpage();
	} else {
		free_page(P);
		if(T->p_pos == T->p_total - 1) p_move(-1);
		else memmove(&T->pages[T->p_pos], &T->pages[T->p_pos + 1], sizeof(Page) * (T->p_total - 1));
		T->p_total -= 1;
	}
  KW_SetLabelText(G.footer->children[0], (P->full[0] != '\0' ? P->full : "empty buffer"));
  close_dropdown();
}

//FILE
#define BUFFER_MAX 4096
static void open_file() {
	char const * selection = tinyfd_openFileDialog( "Select file", "", 0, NULL, NULL, 0 );
	if(selection != NULL) {
		LAST = 0;
		const char* ext 		= strrchr(selection, '.'); //get extension
    char* f;
#ifdef _WIN32
    f = strrchr(selection, '\\');
#else
    f = strrchr(selection, '//');
#endif

		E.buffer_cursor = uvec2is(0);
		E.offset 		    = uvec2is(0);
		E.mark			    = vec2is(-1);

		char* file_name = malloc(strlen(f));
		memcpy(file_name, f + 1, strlen(f));
		file_name[strlen(file_name)] = '\0';
		strtok(file_name, ".");

		init_page(file_name, ext != NULL ? ext : "", f + 1, selection);
    KW_SetLabelText(G.footer->children[0], f + 1);
                
		char* line = malloc(BUFFER_MAX);
    line = memset(line,'\0', BUFFER_MAX); 
		FILE* fp = fopen(selection, "r");
      while (fgets(line, BUFFER_MAX, fp) != 0)  {
        insertRow((line != NULL) ? line : "", P->lines, (strlen(line) > 0) ? strlen(line) : 1);
      }
    P->dirty = 0;
		free(file_name);
		free(line);
		fclose(fp);
	}
	close_dropdown();
}

static void save_file() {
  FILE* fp;
  char* z = NULL;
  if(P->path[0] != '\0') {
    fp = fopen(P->path, "wb");
    if(fp == NULL) return; //no permissions to write to this file, we should dialog this. !!tiny file has this!!
  } else {
    char const * selection = tinyfd_saveFileDialog( "Save File", "", 0, NULL, NULL);
    if(selection == NULL) return;
    const char* ext 		   = strrchr(selection, '.'); //get extension
		const char* f   		   = strrchr(selection, '\\');
    char* file_name        = malloc(strlen(f));
		memcpy(file_name, f + 1, strlen(f));
		file_name[strlen(file_name)] = '\0';
		strtok(file_name, ".");

		edit_page(file_name, ext != NULL ? ext : "", f + 1, selection);

    KW_SetLabelText(G.footer->children[0], f + 1);
    fp = fopen(selection, "wb");
    if(fp == NULL) return;
    free(file_name);
  }
  
  for(uint i = 0; i < P->lines; i++) { //this is seg faulting
    fprintf(fp, "%s\n", P->buffer[i].chars);
  }
  free(z);
  fclose(fp);
  close_dropdown();
}

//WINDOW
static void fetch_scale(SDL_Window* window) {	
	E.visibleW = (E.screenW / (FONT_CHAR_W * E.font_scale)) - 1;
	E.visibleH = (uint)floorf((E.screenH / (FONT_CHAR_H * E.font_scale)) - ((UI_Y_SCALE_BOTTOM + UI_Y_SCALE_TOP) / E.font_scale));						
	E.offset.y = 0; 
	E.offset.x = 0;
}

#if DEBUG
static void test_case() {
  //SHORT LINE
  insertLine(&P->buffer[0], "Hello world, how are you?", 25);
  insertRow("Here is some random lorem ipsum text 5 times!", 1, 45);
  assert(P->buffer[0].size == 25);
  assert(P->buffer[1].size == 45);
  for(uint i = 2; i < 7; i++) {
      insertRow("Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nunc vehicula dui quis felis efficitur dictum. Nam eget tortor vel tellus rhoncus commodo. Class aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos himenaeos. Maecenas ultricies id ex quis varius. Curabitur blandit ligula a nibh rhoncus, rhoncus posuere enim consectetur. Proin at nisi hendrerit, dignissim dui sit amet, aliquam massa. Phasellus posuere quis orci in pharetra. Maecenas ut velit vehicula, tincidunt tellus ac, commodo enim. In et libero convallis, aliquam velit in, euismod quam. Donec non semper libero, quis scelerisque neque. Phasellus quis scelerisque mauris. Nullam et justo dolor. Cras id luctus nunc, eu egestas turpis. Pellentesque ut ullamcorper elit. Ut quis placerat ex, et faucibus ante. Aliquam et condimentum est, sed consequat lacus. Etiam justo quam, varius ut elit vitae, iaculis lacinia lorem. Sed elit ligula, accumsan a varius sed, venenatis vel mauris. Sed faucibus, tortor quis non", i, 1000);
      E.buffer_cursor.y += 1;
      assert(strlen(P->buffer[i].chars) == 1000);
  }
  init_new_tile();
  /* NEW_PAGE(); */
  /* insertLine(&P->buffer[0], "This is a new page....", 22); */
  /* assert(P->buffer[0].size == 22); */
  /* E.buffer_cursor.x = 22; */
}
#endif

static void free_editor() {
  free(E.tiles);
  /* free(E.copy); */
}

int main( int argc, char* args[] )
{
	//sdl setup
	SDL_Window* window     = die_ptr(SDL_CreateWindow("Azul", 0, 0, INIT_SCREEN_WIDTH, INIT_SCREEN_HEIGHT, SDL_WINDOW_RESIZABLE));
	SDL_Renderer* renderer = die_ptr(SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED));
	SDL_SetWindowIcon(window, surface_from_file("logo.png"));

	//Editor init
	E.buffer_cursor = uvec2is(0);
  E.tiles         = malloc(sizeof(TILE) * 1);
  E.current_tile  = 0;
  E.t_total       = 1;
	E.offset 		    = uvec2is(0);
	E.screenW 		  = INIT_SCREEN_WIDTH;
	E.screenH		    = INIT_SCREEN_HEIGHT;
	E.font_scale	  = FONT_SCALE;
	E.visibleW 		  = (INIT_SCREEN_WIDTH  / (FONT_CHAR_W * FONT_SCALE)) - 1;
	E.visibleH		  = ceil((INIT_SCREEN_HEIGHT / (FONT_CHAR_H * FONT_SCALE)) - ((UI_Y_SCALE_BOTTOM + UI_Y_SCALE_TOP) / FONT_SCALE));						
	E.font 			    = font_load_from_file("dos_like_8x14.png", renderer);
	/* E.history		    = init_stack(); */
	E.mark 			    = vec2is(-1);
  //  E.copy          = malloc(COPY_DEFAULT);
  T               = &E.tiles[E.current_tile];

  init_default_tile();
	init_gui(window, renderer);

	NOW      = SDL_GetPerformanceCounter();
	u8 alpha = 255;

	float cursor_timing = 0.0f;

  /* if(DEBUG) test_case(); */

	E.quit = 0;
	u32 frameStart = 0;
	uint frameTime = 0;
	while(E.quit != 1) {
		//Delta time!
		LAST = NOW;
		NOW  = SDL_GetPerformanceCounter();
		deltaTime = (double)((NOW - LAST)*1000 / (double)SDL_GetPerformanceFrequency() );
		frameStart = SDL_GetTicks();
    
		SDL_Event event = {0};
		while(SDL_PollEvent(&event)) {
			switch(event.type) {
      case SDL_QUIT: {				
        E.quit = 1;
      } break;

      case SDL_WINDOWEVENT: {
        case SDL_WINDOWEVENT_SIZE_CHANGED: { 	
          SDL_GetWindowSize(window, &E.screenW, &E.screenH);									
          fetch_scale(window);
          update_gui();
        }break;
      }break;

      case SDL_KEYDOWN: {
        int z = event.key.keysym.sym;
        for(int i = 0; i < sizeof(command_table) / sizeof(command_table[0]); i++) {
          if(command_table[i].key == z) {
            if(ctrl_down && (command_table[i].type != C_NORMAL)) {
              command_table[i].run();
            } else if(ctrl_down != 1 && command_table[i].type != C_CONTROL_KEY) {
              command_table[i].run();
            }
          }				
        }
        cursor_timing = 2.0f; //any key press will "freeze" the cursor
      } break;

      case SDL_KEYUP: {
        switch(event.key.keysym.sym) {
        case SDLK_RCTRL: {
          ctrl_down=0;
        } break;
        case SDLK_LCTRL: {
          ctrl_down=0;
        } break;
        }
      }break;


      case SDL_MOUSEWHEEL: {
        if(!ctrl_down) break;
        if(event.wheel.y > 0) {
          E.font_scale += FONT_SCALE_PER_WHEEL;
          //else PAGE_DOWN();
        } else if (event.wheel.y < 0) {		
          E.font_scale = (E.font_scale <= 0.25) ? 0.25 : E.font_scale - FONT_SCALE_PER_WHEEL;
          //else PAGE_UP();
        }
        fetch_scale(window);					
      } break;

      case SDL_TEXTINPUT: {
        if(ctrl_down) break;
        cursor_timing = 2.0f;
        insertChar(&P->buffer[E.buffer_cursor.y], event.text.text[0]);
      } break;
			}
		}
		die(SDL_SetRenderDrawColor(renderer, UNHEXA(0x000000)));
		die(SDL_RenderClear(renderer));		
		scroll();
		Vec2f pos = vec2f(0, FONT_CHAR_H * FONT_SCALE * UI_Y_SCALE_TOP); //FONT_CHAR_W * FONT_SCALE * 5
		render_text_sized(renderer, &E.font, pos, 0xFFFFFFFFF, E.font_scale, P->buffer);
		render_cursor(renderer, 0xFFFFFFFF, alpha, pos);
    //    render_mark(renderer, 0xFFFFFFFF, 100, pos);

		cursor_timing -= 0.0015f * (float)deltaTime;
		if(cursor_timing <= 0.0f) cursor_timing = 2.0f;

		alpha = (cursor_timing > 0.8) ? 255 : 0;

		render_gui();
		SDL_RenderPresent(renderer);

		frameTime = SDL_GetTicks() - frameStart;
		if(TICKS_PER_FRAME > frameTime) SDL_Delay(TICKS_PER_FRAME - frameTime);
	}

	KW_Quit(G.gui);
	KW_ReleaseSurface(G.driver, G.set);
	KW_ReleaseRenderDriver(G.driver);
	SDL_Quit();
  free_editor();
	return 0;
}
